from typing import Any, Iterable, List, Optional

from sgqlc.endpoint.requests import RequestsEndpoint

from contxt.generated.nionic_queries import Operations as nionic_operations
from contxt.generated.nionic_schema import Facility, MainService, MetricLabel
from contxt.models.ems import ResourceType
from contxt.models.iot import Window
from contxt.services.api import ApiEnvironment, BaseGraphService

from ..auth import Auth
from .pagination import DataPoint, PagedGQLTimeSeries


class NionicService(BaseGraphService):
    """Nionic API client"""

    _envs = (
        ApiEnvironment(
            name="production",
            base_url="https://{tenant}.api.ndustrial.io",
            client_id="vtiZlMRo4apDvThTRiH7kLifQXWUdi9j",
        ),
        ApiEnvironment(
            name="staging",
            base_url="https://{tenant}.api.staging.ndustrial.io",
            client_id="vhGxildn8hRRWZj49y18BGtbjTkFHcTG",
        ),
    )

    def __init__(self, auth: Auth, org_slug: str, env: str = "production", **kwargs) -> None:
        super().__init__(env=env, auth=auth, **kwargs)
        self.base_url = self.base_url.format(tenant=org_slug)
        self.endpoint = RequestsEndpoint(self.base_url.rstrip("/") + "/graphql", session=self.session)

    def get_facilities(self) -> List[Facility]:
        op = nionic_operations.query.facilities
        return (op + self.run(op)).facilities.nodes

    def create_facility(self, data: dict) -> Facility:
        op = nionic_operations.mutation.create_facility
        return self.run(op, data)["data"]["createFacility"]["facility"]

    def get_metric_labels(self) -> List[MetricLabel]:
        op = nionic_operations.query.metric_labels
        return self.run(op).get("data").get("metricLabels")

    def get_metric_data(self, label: str, facility_id: int, start: str) -> List[Any]:
        query = """
        query metricData($facilityId: Int!, $label: String!, $from: String!) {
            facility(id: $facilityId) {
                metricData(label: $label, from: $from) {
                    nodes {
                        time
                        data
                    }
                }
            }
        }
        """
        resp = self.query(query, {"facilityId": int(facility_id), "label": label, "from": start})
        return resp["facility"]["metricData"]["nodes"]

    def get_main_services(
        self, facility_id: int, resource_type: Optional[ResourceType] = None
    ) -> List[MainService]:
        op = nionic_operations.query.main_services
        filterValues = {"facilityId": int(facility_id)}
        return [
            x
            for x in (op + self.run(op, variables=filterValues)).main_services.nodes
            if resource_type is None or x.type.lower() == resource_type.value
        ]

    def get_data_point_data(
        self, name: str, data_source_name: str, start: str, end: str, window: Window, per_page: int
    ) -> Iterable[DataPoint]:
        return PagedGQLTimeSeries(
            api=self,
            query="""
            query dataPointData($dataSourceName: String!, $name: String!
                , $from: String!, $to: String!, $window: String!, $samplingWindow: String!
                , $limit: Int!, $after: Cursor) {
                dataPoint(dataSourceName: $dataSourceName, name: $name) {
                    data(from: $from, to: $to, orderBy: TIME_DESC
                    , window: $window, samplingWindow: $samplingWindow
                    , first: $limit, after: $after) {
                        nodes {
                            time
                            data
                        }
                        pageInfo {
                            hasNextPage
                            endCursor
                        }
                    }
                }
            }
            """,
            params={
                "dataSourceName": data_source_name,
                "name": name,
                "from": start,
                "to": end,
                "window": f"{window.value} seconds",
                # limit to 1 day if window is greater than 1 day
                "samplingWindow": f"{min(window.value, 86400)} seconds",
            },
            per_page=per_page,
        )



from copy import deepcopy
from datetime import datetime, timezone
from typing import Dict, List

from ..utils import is_datetime_aware, make_logger
from .api import ApiEnvironment, ConfiguredApi

logger = make_logger(__name__)


class NgestService(ConfiguredApi):
    """Ngest API client"""

    _envs = (
        ApiEnvironment(
            name="production",
            base_url="https://{tenant}.api.ndustrial.io/rest/data/v1",
            client_id="AhVAWkq2FEoQtWAP7EidZ9uzrc4ED1Dx",
        ),
    )

    def __init__(self, org_slug: str, env: str = "production", **kwargs) -> None:
        self.env = env
        super().__init__(env=env, **kwargs)
        self.base_url = self.base_url.format(tenant=org_slug)

    @staticmethod
    def specialize(
        feed_key: str, feed_token: str, org_slug: str, env: str = "production"
    ) -> "SpecializedNgestService":
        return SpecializedNgestService(
            env=env, feed_key=feed_key, feed_token=feed_token, org_slug=org_slug
        )

    def _format_time_series(self, feed_key: str, time_series: Dict[str, Dict[datetime, float]]) -> Dict:
        # Enforce all datetimes are tz-aware
        assert all(
            is_datetime_aware(dt) for time_series in time_series.values() for dt in time_series.keys()
        ), "Timezone-aware datetimes required"

        # Format for request
        return {
            "feedKey": feed_key,
            "type": "timeseries",
            "data": [
                {
                    "timestamp": dt.astimezone(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
                    "data": {field_descriptor: {"value": str(value)}},
                }
                for field_descriptor, series in time_series.items()
                for dt, value in series.items()
            ],
        }

    def send_time_series(
        self,
        feed_key: str,
        feed_token: str,
        time_series: Dict[str, Dict[datetime, float]],
        per_request: int = 50,
    ) -> List[Dict]:
        """Sends time series data for field(s).

        :param feed_key: feed's key
        :type feed_key: str
        :param feed_token: feed's token
        :type feed_token: str
        :param time_series: dictionary of field descriptors to a time series dictionary
        :type time_series: Dict[str, Dict[datetime, float]]
        :param per_request: number of datapoints to send per request, defaults to 50
        :type per_request: int, optional
        :return: responses
        :rtype: List[Dict]
        """
        # Prepare data for request
        data = self._format_time_series(feed_key=feed_key, time_series=time_series)

        # Send the time series in chunks, partitioned by `per_request`
        responses = []
        while data["data"]:
            # Create chunk for next request
            chunk = deepcopy(data)
            chunk["data"] = data["data"][:per_request]

            # Make request
            response = self.post(f"{feed_token}/ngest/{feed_key}", json=chunk)
            if response["status"] != "ok":
                logger.warning(f"{self.__class__.__name__}: got status {response['status']}")
            responses.append(response)

            # Update remaining data
            data["data"] = data["data"][per_request:]

        return responses


class SpecializedNgestService(NgestService):
    def __init__(self, feed_key: str, feed_token: str, org_slug: str, env: str = "production") -> None:
        super().__init__(env=env, org_slug=org_slug)
        self.feed_key = feed_key
        self.feed_token = feed_token

    def send_time_series(  # type: ignore
        self, time_series: Dict[str, Dict[datetime, float]], per_request: int = 50
    ) -> List[Dict]:
        return super().send_time_series(
            feed_key=self.feed_key,
            feed_token=self.feed_token,
            time_series=time_series,
            per_request=per_request,
        )



from typing import Dict, List, Optional

from ..auth import Auth
from ..models.bus import Channel, ChannelStats
from .api import ConfiguredApi


class MessageBusService(ConfiguredApi):
    """Message Bus API client"""

    _envs = ()

    def __init__(self, auth: Auth, organization_id: str, env: str = "production", **kwargs) -> None:
        super().__init__(env=env, auth=auth, **kwargs)
        self.organization_id = organization_id

    def _channels_url(self, service_id: str) -> str:
        return f"organizations/{self.organization_id}/services/{service_id}/channels"

    def get_channel_for_service(self, channel_id: str, service_id: str) -> Channel:
        resp = self.get(f"{self._channels_url(service_id)}/{channel_id}")
        return Channel.from_api(resp)

    def create_channel_for_service(self, channel: Channel) -> Channel:
        data = {"name": channel.name}
        resp = self.post(f"{self._channels_url(channel.service_id)}", json=data)
        return Channel.from_api(resp)

    def get_channel_with_name_for_service(self, channel_name: str, service_id: str) -> Optional[Channel]:
        for channel in self.get_channels_for_service(service_id):
            if channel.name.lower() == channel_name.lower():
                return channel
        return None

    def get_channels_for_service(self, service_id: str) -> List[Channel]:
        resp = self.get(self._channels_url(service_id))
        return [Channel.from_api(rec) for rec in resp]

    def get_schema_for_channel_and_service(
        self, schema_id: str, channel_id: str, service_id: str
    ) -> Dict:
        # TODO: create model for this response
        return self.get(f"{self._channels_url(service_id)}/{channel_id}/schemas")

    def get_schemas_for_channel_and_service(self, channel_id: str, service_id: str) -> List[Dict]:
        # TODO: create model for this response
        return self.get(f"{self._channels_url(service_id)}/{channel_id}/schemas")  # type: ignore

    def get_stats_for_channel_and_service(self, channel_id: str, service_id: str) -> ChannelStats:
        resp = self.get(f"{self._channels_url(service_id)}/{channel_id}/statistics")
        return ChannelStats.from_api(resp)

    def peek_messages_for_subscription_and_channel_and_service(
        self, subscription: str, channel_id: str, service_id: str
    ) -> Dict:
        return self.get(f"{self._channels_url(service_id)}/{channel_id}/peek/{subscription}")



"""API clients"""

# flake8: noqa
from .auth import AuthService
from .bus import MessageBusService
from .contxt import ContxtService
from .contxt_access import ContxtAccessService
from .contxt_deployments import ContxtDeploymentService
from .ems import EmsService
from .events import EventsService
from .iot import IotDataService, IotService
from .ngest import NgestService
from .nionic import NionicService



from collections import defaultdict
from copy import deepcopy
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from requests import Request

from ..auth import Auth
from ..models import Parsers
from ..models.iot import (
    BatchRequest,
    BatchRequests,
    BatchResponses,
    Feed,
    Field,
    FieldCategory,
    FieldGrouping,
    FieldTimeSeries,
    UnprovisionedField,
    Window,
)
from ..utils import is_datetime_aware, make_logger
from ..utils.object_mapper import ObjectMapper
from .api import ApiEnvironment, ConfiguredApi
from .pagination import DataPoint, PagedRecords, PagedTimeSeries, PageOptions

logger = make_logger(__name__)


class IotService(ConfiguredApi):
    """IOT API client

    Terminology
        - Feed: Data source (i.e. utility meter) with a set of fields
        - Field: A specific time series data from a feed
        - Grouping: Group of fields
    """

    _envs = (
        ApiEnvironment(
            name="production",
            base_url="https://{tenant}.api.ndustrial.io/rest/iot/v1",
            client_id="iznTb30Sfp2Jpaf398I5DN6MyPuDCftA",
        ),
        ApiEnvironment(
            name="staging",
            base_url="https://{tenant}.api.staging.ndustrial.io/rest/iot/v1",
            client_id="m35AEcxD8hf65sq04ZU7yFxqpqVkKzES",
        ),
    )

    def __init__(self, auth: Auth, org_slug: str, env: str = "production", **kwargs) -> None:
        super().__init__(env=env, auth=auth, **kwargs)
        self.base_url = self.base_url.format(tenant=org_slug)

    def provision_field_for_feed(self, feed_id: int, field: Field) -> Field:
        resp = self.post(f"feeds/{feed_id}/fields", data=field.post())
        return Field.from_api(resp)

    def create_grouping(
        self, facility_id: int, label: str, description: str, is_public: bool, field_category_id: str
    ) -> FieldGrouping:
        res = self.post(
            f"facilities/{facility_id}/groupings",
            json={
                "label": label,
                "description": description,
                "is_public": is_public,
                "fileds": field_category_id,
            },
        )
        return FieldGrouping.from_api(res)

    def add_field_to_grouping(self, grouping_id: int, field_id: int):
        return self.post(f"groupings/{grouping_id}/fields/{field_id}")

    def set_fields_for_grouping(self, grouping_id: str, fields: List[str]):
        return self.post(f"groupings/{grouping_id}/fields", json={"fields": fields})

    def unprovision_field(self, field_id: int):
        return self.delete(uri=f"fields/{field_id}")

    def delete_field_grouping(self, grouping_id: str):
        return self.delete(uri=f"groupings/{grouping_id}")

    def delete_time_series_point(
        self, output_id: int, field: str, interval: Window, time: datetime
    ) -> Optional[Dict]:
        """Delete a point from the time series data by interval and timestamp"""
        return self.delete(
            f"outputs/{output_id}/fields/{field}/data/window/{interval.value}/event_time/{time}"
        )

    def get_feed_with_id(self, id: int) -> Feed:
        """Get feed with id `id`"""
        return Feed.from_api(self.get(f"feeds/{id}"))

    def get_feed_with_key(self, key: str) -> Optional[Feed]:
        """Get feed with key `key`"""
        feeds = self.get_feeds(key=key)
        N = len(feeds)  # type: ignore
        if N == 0:
            return None
        elif N == 1:
            return feeds[0]  # type: ignore
        raise KeyError(f"Expected singleton feed with key {key}, not {N}")

    def update_feed_provisioning_mode(self, id: int, enabled: bool):
        """Enable/Disable provisioning mode for a feed"""
        return self.put(uri=f"feeds/{id}", json={"provisioning_mode": enabled})

    def get_feeds(
        self,
        facility_id: Optional[int] = None,
        key: Optional[str] = None,
        page_options: Optional[PageOptions] = None,
    ) -> Iterable[Feed]:
        """Get feeds with facility id `facility_id` and/or key `key`"""
        return PagedRecords(
            api=self,
            url="feeds",
            params={"facility_id": facility_id, "key": key},
            options=page_options,
            record_parser=Feed.from_api,
        )

    def get_fields_for_facility(
        self, facility_id: int, page_options: Optional[PageOptions] = None
    ) -> Iterable[Field]:
        """Get fields for facility with id `facility_id`"""
        return PagedRecords(
            api=self,
            url=f"facilities/{facility_id}/fields",
            options=page_options,
            record_parser=Field.from_api,
        )

    def get_fields_for_feed(
        self, feed_id: int, page_options: Optional[PageOptions] = None
    ) -> Iterable[Field]:
        """Get fields for feed with id `feed_id`"""
        return PagedRecords(
            api=self, url=f"feeds/{feed_id}/fields", options=page_options, record_parser=Field.from_api
        )

    def create_field(self, field: Field) -> Field:
        """Create field"""
        data = field.post()
        resp = self.post(f"outputs/{field.output_id}/fields", data=data)
        return Field.from_api(resp)

    def get_time_series_for_field(
        self,
        field: Field,
        start_time: Union[datetime, None] = None,
        window: Window = Window.RAW,
        end_time: Optional[datetime] = None,
        per_page: int = 1000,
    ) -> Iterable[DataPoint]:
        """Get time series data for field `field`"""
        # Manually validate the window choice, since our API does not return a
        # helpful error message
        assert isinstance(window, Window), "window must be of type Window"
        assert (start_time is None) == (
            end_time is None
        ), "Either both start and end time should be provided, or both should be missing"
        return PagedTimeSeries(
            api=self,
            url=f"outputs/{field.output_id}/fields/{field.field_human_name}/data",
            params={
                "timeStart": int(start_time.timestamp()) if start_time else None,
                "timeEnd": int(end_time.timestamp()) if end_time else None,
                "window": window.value,
            },
            per_page=per_page,
        )

    def get_time_series_for_fields(
        self,
        fields: List[Field],
        start_time: Union[datetime, None] = None,
        window: Window = Window.RAW,
        end_time: Optional[datetime] = None,
    ) -> List[FieldTimeSeries]:
        """Get complete (non-paginated) time series data for each field in `fields`"""
        assert (start_time is None) == (
            end_time is None
        ), "Either both start and end time should be provided, or both should be missing"
        # Build requests queue
        params = {
            "timeStart": int(start_time.timestamp()) if start_time else None,
            "timeEnd": int(end_time.timestamp()) if end_time else None,
            "window": window.value,
            "limit": 5000,
        }
        queue: List[Tuple[str, BatchRequest]] = [
            (
                f.field_human_name,  # type: ignore
                BatchRequest.from_request(
                    Request(
                        method="GET",
                        url=self._url(f"outputs/{f.output_id}/fields/{f.field_human_name}/data"),
                        params=params,
                    )
                ),
            )
            for f in fields
        ]

        # Make all requests in queue
        MAX_BATCH_REQUESTS = 200
        records: Dict[str, Dict[datetime, str]] = defaultdict(dict)
        while queue:
            # Make next batch of requests
            requests = dict(queue[:MAX_BATCH_REQUESTS])
            del queue[:MAX_BATCH_REQUESTS]
            logger.info(f"Making {len(requests)} batched requests to IOT API")
            responses = self._batch_request(requests)
            any_success = False

            # Process responses
            for name, resp in responses.items():
                if resp.ok:
                    any_success = True
                    series = {
                        Parsers.datetime(r["event_time"]): Parsers.unknown(r["value"])
                        for r in resp.body["records"]
                    }
                    records[name].update(series)
                    # Add request for next page
                    next_page_url = resp.body["meta"]["next_page_url"]
                    if next_page_url:
                        queue.append((name, BatchRequest(method="GET", uri=next_page_url)))
                else:
                    # Retry request
                    logger.warning(f"Got bad response from IOT API ({resp.body}). Retrying...")
                    queue.append((name, requests[name]))

            if not any_success:
                raise OSError(f"All {len(requests)} batched requests to IOT API failed")

        fields_by_name = {f.field_human_name: f for f in fields}
        return [
            FieldTimeSeries(field=fields_by_name[name], time_series=series)
            for name, series in records.items()
        ]

    def get_time_series_for_field_grouping(
        self, grouping_id: str, **kwargs
    ) -> List[Iterable[DataPoint]]:
        """Get time series data for fields in grouping with id `grouping_id`"""
        grouping = self.get_field_grouping(grouping_id)
        return [self.get_time_series_for_field(field=f, **kwargs) for f in grouping.fields]

    def get_unprovisioned_fields_for_feed_id(self, feed_id: int) -> List[UnprovisionedField]:
        """Get unprovisioned fields for feed with id `feed_id`"""
        return [
            UnprovisionedField.from_api(rec) for rec in self.get(f"feeds/{feed_id}/fields/unprovisioned")
        ]

    def get_unprovisioned_fields_for_feed_key(self, feed_key: str) -> Optional[List[UnprovisionedField]]:
        """Get unprovisioned fields for feed with key `feed_key`"""
        feed = self.get_feed_with_key(key=feed_key)
        if not feed:
            return None
        return [
            UnprovisionedField.from_api(rec) for rec in self.get(f"feeds/{feed.id}/fields/unprovisioned")
        ]

    def get_field_grouping(self, id: str) -> FieldGrouping:
        """Get field grouping with id `id`"""
        return FieldGrouping.from_api(self.get(f"groupings/{id}"))

    def get_categories_for_facility(self, facility_id: int) -> List[FieldCategory]:
        """Get categories for facility with id `facility_id`"""
        res = [FieldCategory.from_api(rec) for rec in self.get(f"facilities/{facility_id}/categories")]
        categories = []
        for cat in res:
            categories.append(cat)
            for sub in cat.subcategories:
                categories.append(sub)
        return categories

    def add_grouping_to_category(self, grouping_id: str, field_category_id: str):
        return self.put(
            f"groupings/{grouping_id}",
            json={
                "field_category_id": field_category_id,
            },
        )

    def get_field_groupings_for_facility(
        self, facility_id: int, page_options: Optional[PageOptions] = None
    ) -> Iterable[FieldGrouping]:
        """Get field groupings for facility with id `facility_id`"""
        return PagedRecords(
            api=self,
            url=f"facilities/{facility_id}/groupings",
            options=page_options,
            record_parser=FieldGrouping.from_api,
        )

    def _batch_request(self, requests: BatchRequests) -> BatchResponses:
        prepared_requests = {label: req.to_api() for label, req in requests.items()}
        resp = self.post("batch", json=prepared_requests)
        return ObjectMapper.tree_to_object(resp, BatchResponses)


def format_time_series(feed_key: str, time_series: Dict[str, Dict[datetime, float]]) -> Dict:
    # Enforce all datetimes are tz-aware
    assert all(
        is_datetime_aware(dt) for time_series in time_series.values() for dt in time_series.keys()
    ), "Timezone-aware datetimes required"

    # Format for request
    return {
        "feedKey": feed_key,
        "type": "timeseries",
        "data": [
            {
                "timestamp": dt.astimezone(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
                "data": {field_descriptor: {"value": str(value)}},
            }
            for field_descriptor, series in time_series.items()
            for dt, value in series.items()
        ],
    }


NgestField = str
NgestRecord = Tuple[datetime, Dict[NgestField, Any]]


class IotDataService(ConfiguredApi):
    """IOT API client v2.0

    Terminology
        - Feed: Data source (i.e. utility meter) with a set of fields
        - Field: A specific time series data from a feed
        - Grouping: Group of fields
    """

    _envs = (
        ApiEnvironment(
            name="production",
            base_url="https://iot.api.ndustrial.io/v2/",
            client_id="ZPrYMWVCcsyYaKKK2uiFLS71X1MB7zJP",
        ),
        ApiEnvironment(
            name="development",
            base_url="http://localhost:8080/v2/",
            client_id="ZPrYMWVCcsyYaKKK2uiFLS71X1MB7zJP",
        ),
    )

    def __init__(self, org_id: str, auth: Auth, env: str = "production", **kwargs) -> None:
        super().__init__(env=env, auth=auth, **kwargs)
        self.org_id = org_id

    def get_source_data(
        self, source_key: str, start: datetime, resolution: timedelta, end: Optional[datetime] = None
    ):
        """Get the data for the given source"""
        assert start.tzinfo
        assert resolution.total_seconds() == int(resolution.total_seconds())
        params = {
            "start": int(start.timestamp()),
            "end": int(end.timestamp()) if end else int(datetime.now().timestamp()),
            "resolution": f"P{resolution.days}DT{resolution.seconds}S",
        }
        return self.get(f"org/{self.org_id}/sources/{source_key}/data", params=params)

    def get_source_cursor(self, source_key: str) -> Optional[datetime]:
        """Get the cursor for the given source"""
        body = self.get(f"org/{self.org_id}/sources/{source_key}/cursor")
        assert (
            body["source_key"] == source_key
        ), f"Got unexpected source key on response. Requested {source_key}, but got {body['source_key']}"
        epoch = body["cursor_epoch"]
        if epoch:
            return datetime.fromtimestamp(epoch, tz=timezone.utc)
        return None

    def ingest_source_data(
        self, source_key: str, data: List[NgestRecord], batch_size: int = 50
    ) -> List[Dict]:
        responses = []
        tail = data
        while tail:
            batch, tail = tail[:batch_size], tail[batch_size:]
            _data = []
            for record in batch:
                dt, field_values = record
                assert is_datetime_aware(dt), f"Ngest requires timezone-aware datetimes, got: {dt}"
                _data.append(
                    {
                        "timestamp": dt.astimezone(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
                        "data": {k: {"value": str(v)} for k, v in field_values.items()},
                    }
                )
            msg = {"feedKey": source_key, "type": "timeseries", "data": _data}

            # Make request
            response = self.post(f"org/{self.org_id}/ngest/{source_key}", json=msg)
            if response["status"] != "ok":
                logger.warning(f"{self.__class__.__name__}: got status {response['status']}")
            responses.append(response)

        return responses

    # fixme: deprecated
    def send_time_series(
        self, feed_key: str, time_series: Dict[str, Dict[datetime, float]], per_request: int = 50
    ) -> List[Dict]:
        """
        This method is deprecated, use ingest_source_data instead

        Sends time series data for field(s).

        :param feed_key: feed's key
        :type feed_key: str
        :param time_series: dictionary of field descriptors to a time series dictionary
        :type time_series: Dict[str, Dict[datetime, float]]
        :param per_request: number of datapoints to send per request, defaults to 50
        :type per_request: int, optional
        :return: responses
        :rtype: List[Dict]
        """
        # Prepare data for request
        data = format_time_series(feed_key=feed_key, time_series=time_series)

        # Send the time series in chunks, partitioned by `per_request`
        responses = []
        while data["data"]:
            # Create chunk for next request
            chunk = deepcopy(data)
            chunk["data"] = data["data"][:per_request]

            # Make request
            response = self.post(f"org/{self.org_id}/ngest/{feed_key}", json=chunk)
            if response["status"] != "ok":
                logger.warning(f"{self.__class__.__name__}: got status {response['status']}")
            responses.append(response)

            # Update remaining data
            data["data"] = data["data"][per_request:]

        return responses

    def get_source_field_cursor(
        self, source_key: str, field_name: Union[str, None] = None
    ) -> Optional[datetime]:
        """Get the cursor for the given source and field"""

        url = f"org/{self.org_id}/sources/{source_key}"
        if field_name is not None:
            url += f"/fields/{field_name}"
        url += "/cursor"

        body = self.get(url)
        assert (
            body["source_key"] == source_key
        ), f"Got unexpected source key on response. Requested {source_key}, but got {body['source_key']}"
        epoch = body["cursor_epoch"]
        if epoch:
            return datetime.fromtimestamp(epoch, tz=timezone.utc)
        return None



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, FrozenSet, Optional, Tuple

import requests
from requests import PreparedRequest, Response, Session
from requests.adapters import HTTPAdapter
from requests.auth import AuthBase
from requests.exceptions import HTTPError
from sgqlc.endpoint.requests import RequestsEndpoint
from sgqlc.operation import Operation
from urllib3.util.retry import Retry

from contxt import __version__

from ..auth import Auth, TokenProvider
from ..utils import make_logger

logger = make_logger(__name__)

# NOTE: support breaking change in urllib3: https://github.com/urllib3/urllib3/issues/2092
RETRY_METHODS_PARM = "allowed_methods" if hasattr(Retry(), "allowed_methods") else "method_whitelist"


class BearerTokenAuth(AuthBase):
    """Bearer token to authorize requests"""

    def __init__(self, token_provider: TokenProvider) -> None:
        self.token_provider = token_provider

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        request.headers["Authorization"] = f"Bearer {self.token_provider.access_token}"
        return request


class ApiRetry(Retry):
    def __init__(
        self,
        total: int = 3,
        backoff_factor: float = 0.1,
        method_whitelist: FrozenSet = frozenset(
            # NOTE: be careful as some of these methods are not idempotent
            ["DELETE", "GET", "OPTIONS", "POST", "PUT", "TRACE", "HEAD"]
        ),
        status_forcelist: Tuple[int, ...] = (500, 502, 504),
        raise_on_status: bool = False,
        **kwargs,
    ) -> None:
        kwargs[RETRY_METHODS_PARM] = method_whitelist
        super().__init__(
            total=total,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
            raise_on_status=raise_on_status,
            **kwargs,
        )


class Api:
    """An API with url `base_url`.

    If `token_provider` is specified, all requests will be authenticated with
    the access token it provides.
    """

    def __init__(
        self,
        base_url: str,
        token_provider: Optional[TokenProvider] = None,
        retry: Optional[ApiRetry] = ApiRetry(),
    ) -> None:
        self.base_url = base_url if base_url.endswith("/") else f"{base_url}/"

        # Initialize session
        self.session = Session()
        self.session.auth = BearerTokenAuth(token_provider) if token_provider else None
        self.session.headers.update({"Cache-Control": "no-cache"})
        self.session.headers.update({"User-Agent": f"contxt-sdk-python/{__version__}"})
        self.session.hooks = {"response": self._log_response}  # type: ignore

        # Attach retry adapter
        if retry:
            adapter = HTTPAdapter(max_retries=retry)
            self.session.mount("http://", adapter)
            self.session.mount("https://", adapter)

    def _url(self, uri: str) -> str:
        return f"{self.base_url}{uri}"

    def _log_response(self, response: Response, *args, **kwargs) -> None:
        t = response.elapsed.total_seconds()
        logger.debug(
            f"Called {response.request.method} {response.url} with body"  # type: ignore
            f" {response.request.body!r} ({t} s)"
        )

    def _process_response(self, response: Response) -> Dict:
        try:
            # Raise any error
            response.raise_for_status()
        except HTTPError:
            # Catch the error, to log the response's message, and reraise
            # Try to decode the response as json, else fall back to raw text
            response_json = self._get_json(response)
            msg = (
                response_json.get("message")
                if isinstance(response_json, dict)
                else response_json or response.text
            )
            logger.debug(f"HTTP Error: {response.reason} - {msg}")
            raise

        # Return json, if any
        return self._get_json(response)

    def _get_json(self, response: Response) -> Dict:
        try:
            return response.json()
        except ValueError:
            return {}

    def get(self, uri: str, params: Optional[Dict] = None, **kwargs) -> Dict:
        """Sends a GET request"""
        response = self.session.get(url=self._url(uri), params=params, **kwargs)
        return self._process_response(response)

    def raw_post(
        self, uri: str, data: Optional[Dict] = None, json: Optional[Dict] = None, **kwargs
    ) -> requests.Response:
        """Sends a POST request without processing response"""
        return self.session.post(url=self._url(uri), data=data, json=json, **kwargs)

    def post(self, uri: str, data: Optional[Dict] = None, json: Optional[Dict] = None, **kwargs) -> Dict:
        """Sends a POST request"""
        response = self.session.post(url=self._url(uri), data=data, json=json, **kwargs)
        return self._process_response(response)

    def put(self, uri: str, data: Optional[Dict] = None, json: Optional[Dict] = None, **kwargs) -> Dict:
        """Sends a PUT request"""
        response = self.session.put(url=self._url(uri), data=data, json=json, **kwargs)
        return self._process_response(response)

    def delete(self, uri: str, **kwargs) -> Dict:
        """Sends a DELETE request"""
        response = self.session.delete(url=self._url(uri), **kwargs)
        return self._process_response(response)


@dataclass
class ApiEnvironment:
    """An environment for an API.

    Note `client_id` is only needed if authentication is required.
    """

    name: str
    base_url: str
    client_id: str


class ConfiguredApi(Api, ABC):
    """An `Api` configured for multiple environments, such as staging and production.
    Available environments are expressed by `_envs` and the desired environment is
    specified by its name `env`.

    Overload this class to implement `_envs`.
    """

    def __init__(self, env: str, auth: Optional[Auth] = None, **kwargs) -> None:
        # TODO: figure out a cleaner approach to environment selection
        api_env = self._get_env(env)
        self.env = env
        self.client_id = api_env.client_id
        token_provider = auth.get_token_provider(self.client_id) if auth else None
        super().__init__(base_url=api_env.base_url, token_provider=token_provider, **kwargs)

    @classmethod
    def _get_env(cls, name: str) -> ApiEnvironment:
        """Get environment with name `name`"""
        envs = {e.name: e for e in cls._envs}  # type: ignore
        if name not in envs:
            raise KeyError(f"Invalid environment '{name}'. Choose from {list(envs.keys())}.")
        return envs[name]

    @property  # type: ignore
    @classmethod
    @abstractmethod
    def _envs(cls) -> Tuple[ApiEnvironment, ...]:
        """Lists available environments."""
        pass


class BaseGraphService(ConfiguredApi):
    def __init__(self, auth: Auth, env: str, **kwargs) -> None:
        super().__init__(env, auth, **kwargs)
        self.endpoint = RequestsEndpoint(f"{self.base_url}/graphql", session=self.session)

    def query(self, query: str, variables: Optional[Any] = None) -> Dict:
        """Send a GraphQL query"""
        resp = self.post("graphql", json={"query": query, "variables": variables})
        if "errors" in resp:
            raise Exception(resp["errors"][0]["message"])
        return resp["data"]

    def run(self, op: Operation, variables: Optional[Any] = None) -> Any:
        data = self.endpoint(op, variables)
        if "errors" in data:
            raise Exception(data["errors"][0]["message"])
        return data



from typing import Dict, List, Union

from .api import ApiEnvironment, ConfiguredApi


class AuthService(ConfiguredApi):
    """Auth API client"""

    _envs = (
        ApiEnvironment(
            name="production",
            base_url="https://contxtauth.com/v1",
            client_id="75wT048QcpE7ujwBJPPjr263eTHl4gEX",
        ),
        ApiEnvironment(
            name="staging",
            base_url="https://contxt-auth-service.staging.ndustrial.io/v1",
            client_id="7TceUsM1eC4nKmdoC717383DWyfc9QoY",
        ),
    )

    def __init__(self, env: str = "production", **kwargs) -> None:
        super().__init__(env=env, **kwargs)

    def get_jwks(self) -> Dict:
        return self.get(".well-known/jwks.json")

    def get_token(self, access_token: str, audiences: Union[str, List[str]]) -> Dict:
        audiences = [audiences] if isinstance(audiences, str) else audiences
        return self.post(
            "token", json={"audiences": audiences}, headers={"Authorization": f"Bearer {access_token}"}
        )

    def get_cluster_config(self, access_token: str, host: str) -> Dict:
        return self.post(
            "clusterconfig", headers={"Authorization": f"Bearer {access_token}"}, json={"host": host}
        )

    def get_oauth_token(self, client_id: str, client_secret: str, audience: str) -> Dict:
        return self.post(
            "oauth/token",
            json={
                "client_id": client_id,
                "client_secret": client_secret,
                "audience": audience,
                "grant_type": "client_credentials",
            },
        )



from typing import List, Optional

import click

from contxt.cli.clients import Clients
from contxt.cli.utils import OPTIONAL_PROMPT_KWARGS, fields_option, print_item, print_table, sort_option
from contxt.models.contxt import Project


@click.group()
def projects() -> None:
    """Projects."""


@projects.command()
@click.argument("slug", default="")
@click.pass_obj
@fields_option(default="id, slug, name, type, description", obj=Project)
@sort_option(default="slug")
def get(clients: Clients, slug: Optional[str], fields: List[str], sort: str) -> None:
    """Get project(s)"""
    items = (
        [clients.contxt_deployments.get(f"{clients.org_id}/projects/{slug}")]
        if slug
        else clients.contxt_deployments.get(f"{clients.org_id}/projects")
    )
    print_table(items=items, keys=fields, sort_by=sort)


@projects.command()
@click.option("--name", prompt=True)
@click.option("--slug", **OPTIONAL_PROMPT_KWARGS)
@click.option("--type", type=click.Choice(["Core", "Custom"]), default="Custom", prompt=True)
@click.option("--description", **OPTIONAL_PROMPT_KWARGS)
@click.option("--owner-role-id", **OPTIONAL_PROMPT_KWARGS)
@click.pass_obj
def create(
    clients: Clients,
    name: str,
    description: str,
    type: str,
    slug: Optional[str],
    owner_role_id: Optional[str],
) -> None:
    """Create a project"""
    result = clients.contxt_deployments.post(
        f"{clients.org_id}/projects",
        {
            "name": name,
            "slug": slug,
            "description": description,
            "type": type,
            "owner_role_id": owner_role_id,
        },
    )
    print_item(result)


@projects.command()
@click.argument("curr_slug", metavar="SLUG")
@click.option("--name")
@click.option("--slug")
@click.option("--type")
@click.option("--description")
@click.pass_obj
def update(
    clients: Clients,
    curr_slug: str,
    name: Optional[str],
    slug: Optional[str],
    type: Optional[str],
    description: Optional[str],
) -> None:
    """Update a project"""
    clients.contxt_deployments.put(
        f"{clients.org_id}/projects/{curr_slug}",
        {"name": name, "slug": slug, "type": type, "description": description},
    )


@projects.command()
@click.argument("slug", default="")
@click.pass_obj
def delete(clients: Clients, slug: str) -> None:
    """Delete a project"""
    clients.contxt_deployments.delete(f"{clients.org_id}/projects/{slug}")
